//! # RefCell<T> and the interior mutability pattern
//!
//! Interior mutability allows mutate data when immutable
//! references to said data exists. This is possible due to
//! the unsafe code. Unsafe code more in chapter 20.
//!
//! Can use types that use interior mutability pattern when
//! can ensure borrowing rules will be followed at runtime.
//!
//! ## Enforcing borrowing Rules at Runtime with RefCell<T>
//!
//! RefCell<T> represents single ownership over data it holds.
//! If the Rust compiler can't be sure the code compiles with
//! ownership rules, it might reject a correct program.
//!
//! A recap of reasons to choose Box<T>, Rc<T>, or RefCell<T>:
//! 1) Rc<T> enables multiple owners of the same data; Box<T>
//!    and RefCell<T> have single owners.
//! 2) Box<T> allows immutable or mutable borrows checked at
//!    compile time; Rc<T> allows only immutable borrows checked
//!    at compile time; RefCell<T> allos immutable or mutable
//!    borrows checked at runtime.
//! 3) Because RefCell<T> allows mutable borrows checked at
//!    runtime, you can mutate the value inside the RefCell<T>
//!    even when the RefCell<T> is immutable.
//!
//! Mutating the value inside an immutable value is the interior
//! mutability pattern.

#[derive(Debug)]
enum List {
    Cons(Rc<RefCell<i32>>, Rc<List>),
    Nil,
}

use crate::List::{Cons, Nil};
use std::cell::RefCell;
use std::rc::Rc;

fn main() {
    // Interior mutability: A mutable borrow to an immutable value

    // this panic!s
    // let x = 5;
    // let y = &mut x;

    // A use case for interior mutability: mock objects

    // Sometimes use a type in place of another type, in order
    // to observe particular behavior and assert it's implemented
    // correctly. Called a test double. Stand in for other types
    // when we're running tests.
    //
    // Mock objects are specific types of test doubles that record
    // what happens during a test so can assert correct actions
    // took place.

    // Having multiple owners of mutable data by combining Rc<T> and RefCell<T>
    // Only gives immutable access to data in this fashion.

    let value = Rc::new(RefCell::new(5));

    let a = Rc::new(Cons(Rc::clone(&value), Rc::new(Nil)));

    let b = Cons(Rc::new(RefCell::new(3)), Rc::clone(&a));
    let c = Cons(Rc::new(RefCell::new(4)), Rc::clone(&a));

    *value.borrow_mut() += 10;

    println!("a after = {a:?}");
    println!("b after = {b:?}");
    println!("c after = {c:?}");
}
