//! # Shared-State Concurrency
//!
//! This is another way to handle concurrency. Sharing data.
//! Consider the slogan: "do not communicate by sharing memory."

use std::rc::Rc;
use std::sync::{Arc, Mutex};
use std::thread;

fn main() {
    // Using mutexes to allow access to data from one thread at a time
    //
    // Mutex is an abbreviation for mutual exclusion. Allows only one
    // thread to access some data at any given time.
    // To access data in a mutex, thread must first signal it wants
    // access by asking to acquire the mutex's lock. The lock is
    // a data structure part of the mutex that keeps track of who
    // currently has exclusive access to the data. Therefore,
    // mutex described as guarding the data.
    //
    // Mutexes have two rules:
    // 1) Must attempt to acquire the lock before using the data
    // 2) When done with data that mutex guards, must unlock data
    //    so other threads can acquire lock.

    let m = Mutex::new(5);

    {
        let mut num = m.lock().unwrap();
        *num = 6;
    }

    println!("m = {m:?}");

    // Sharing a Mutex<T> Between multiple threads

    // let counter = Mutex::new(0);
    // let mut handles = vec![];

    // for _ in 0..10 {
    //     let handle = thread::spawn(move || {
    //         let mut num = counter.lock().unwrap();

    //         *num += 1;
    //     });
    //     handles.push(handle);
    // }

    // for handle in handles {
    //     handle.join().unwrap();
    // }

    // println!("Result: {}", *counter.lock().unwrap());

    // Multiple ownership with multiple threads

    // let counter = Rc::new(Mutex::new(0));
    // let mut handles = vec![];

    // for _ in 0..10 {
    //     let counter = Rc::clone(&counter);
    //     let handle = thread::spawn(move || {
    //         let mut num = counter.lock().unwrap();

    //         *num += 1;
    //     });
    //     handles.push(handle);
    // }

    // for handle in handles {
    //     handle.join().unwrap();
    // }

    // println!("Result: {}", *counter.lock().unwrap());

    // Atomic reference counting with Arc<T>
    //
    // Arc<T> is like Rc<T> is safe to use concurrent situations.
    // The a stands for atomic meanings it's an atomically reference
    // counted type. Thread safety like this comes at an example.
    // Arc<T> and Rc<T> have the same API.

    let counter = Arc::new(Mutex::new(0));
    let mut handles = vec![];

    for _ in 0..10 {
        let counter = Arc::clone(&counter);
        let handle = thread::spawn(move || {
            let mut num = counter.lock().unwrap();

            *num += 1;
        });
        handles.push(handle);
    }

    for handle in handles {
        handle.join().unwrap();
    }

    println!("Result: {}", *counter.lock().unwrap());
}
