//! Futures and the Async Syntax
//!
//! Key elements of async are *futures*, `async`, and `await`.
//! A *future* is a value which may not be ready now, but will
//! be ready at some point in the future.
//!
//! The `async` keyword can be applied to blocks and functions
//! to specify they can be interrupted and resumed. `await`
//! keyword to wait for a future to become ready.
//!
//! Rust handles these keywords differently from languages
//! that have this feature already.

use std::future::Future;
use trpl::Html;

/// Writing `async fn` is equivalent to writing a function which
/// returns a future of the return type.
async fn page_title(url: &str) -> Option<String> {
    let response = trpl::get(url).await.text().await;
    Html::parse(&response_text)
        .select_first("title")
        .map(|title_element| title_element.inner_html())
}

/// This function is equivalent to the function above.
fn page_title(url: &str) -> impl Future<Output = Option<String>> + '_ {
    async move {
        let text = trpl::get(url).await.text().await;
        Html::parse(&text)
            .select_first("title")
            .map(|title| title.inner_html())
    }
}

/// Rust doesn't allow for main to be async
/// This can't happen because async code needs a runtime: a Rust crate
/// which manages the details of executing async code. `main` can
/// initialize a runime but isn't a runtime itself.
///
/// There are also different versions of async runtimes tailored to
/// specific needs. For example, a high-throughput web server with many
/// different CPU cores and a lot of RAM has different needs than a
/// microcontroller with 1 core, a small amount of RAM, and no ability
/// to do heap allocations.
fn main() {
    let args: Vec<String> = std::env::args().collect();

    trpl::run(async {
        let url = &args[1];
        match page_title(url).await {
            Some(title) => println!("The title for {url} is {title}"),
            None => println!("{url} had no title"),
        }
    })
}
